 
// Uncomment these defines to show extra debug information
#define DEBUG_POST
//#define DEBUG_RAM
#define DEBUG_PAGES
//#define DEBUG_ACTIONS
//#define DEBUG_DEVICES
//#define DEBUG_EVENTS

// Libraries to handle all the hardware
#include "Utils.h"
#include "SPI.h"
#include "Wire.h"
#include "Ethernet.h"
#include "WebServer2.h"
#include "Wire.h"
#include "I2C_eeprom.h"
#include "Clock.h"
#include "ChaconController.h"
#include "ShiftRegister.h"
#include "EEPROM.h"

// Pin connected to the test button (currently not in use)
// #define INPUT_PIN 8

// Initializes the shift register chip(s)
// Parameters are  (SER_pin  RCLK_pin  SRCLK_pin  number_of_outputs ) 
ShiftRegister sr = ShiftRegister(7, 6, 5, 16);

// Initializes tha class that communicated with the external EEPROM
// Parameters are the address of the first chip, and the capacity of each chip
// The capacity is in Kbytes, NOT Kbits
I2C_eeprom flash(0x50, 32768);

// Libraries that contain the data structures
#include "OrderedList.h"
#include "Device.h"
#include "Schedule.h"
#include "Event.h"

// Dimensions for the different images than appear on the pages
#define BLUEPRINT_WIDTH 360
#define BLUEPRINT_HEIGHT 336
#define DEVICE_IMAGE_DIMENSION 30

// Network Settings
static uint8_t mac[] = { 0x90, 0xA2, 0xDA, 0x00, 0x75, 0x57 };
static uint8_t ip[] = { 192, 168, 1, 69 };

// Password to acess the system
#define SYSTEM_PASSWORD "bananas"

// Parameter names used in the POST requests
// Using only one char to keep the buffers small and save memory
#define PARAM_TIME     't'    
#define PARAM_DEVICE   'd'
#define PARAM_REMOVE   'v'
#define PARAM_RESET    'r'
#define PARAM_ADD      'n'
#define PARAM_MAP      'm'
#define PARAM_DATA_1   'x'
#define PARAM_SAVE     's'
#define PARAM_PASSWORD 'p'
#define PARAM_LOGOUT   'l'
#define PARAM_CLOCK    'c'    

// Names of the webpages thet will be created
#define PREFIX ""
//#define PAGE_DIRECT "direct"
//#define PAGE_SCHEDULER "scheduler"
//#define PAGE_BLUEPRINT "blueprint"
//#define PAGE_LOGIN "login"

char* pageNames[]={"direct", "scheduler", "blueprint", "login"};
#define PAGE_NUMBER_DIRECT     0
#define PAGE_NUMBER_SCHEDULER  1
#define PAGE_NUMBER_BLUEPRINT  2
#define PAGE_NUMBER_LOGIN      3

// Defines to make using images easiear. All the images are numbered, and the
// order must be sequencial and match the files that were uploaded to the EEproms
#define IMAGE_ICON            1
#define IMAGE_SWITCH          2
#define IMAGE_CLOCK           3
#define IMAGE_HOUSE           4
#define IMAGE_POWER           5
#define IMAGE_PLANT           6
#define IMAGE_CANCEL          7
#define IMAGE_ADD             8
#define IMAGE_SAVE            9
#define IMAGE_LIGHT_ON        10
#define IMAGE_LIGHT_OFF       11
#define IMAGE_WINDOW_OPEN     12
#define IMAGE_WINDOW_CLOSED   13
#define IMAGE_KEY_ON          14
#define IMAGE_KEY_OFF         15
#define IMAGE_OUTLET_ON       16
#define IMAGE_OUTLET_OFF      17
#define IMAGE_XMAS_ON         18
#define IMAGE_XMAS_OFF        19
#define IMAGE_CONFIRM         20
#define TOTAL_IMAGES          20
#define IMAGE_NO_NUMBER       255

// Generic buffer used in int<->string conversions
char buff[16];

// Variables to process POST parameters
#define NAMELEN 8
#define VALUELEN 16
char name[NAMELEN];
char value[VALUELEN];

// Webserver that will handle all the http requests
WebServer webserver;

// Common strings stored in progmem. These are used to construct the webpages
P(newLine) = "<br>\n";
P(strCenterBegin) = "<center>";
P(beginTableRow) = "<tr>\n<td>";
P(middleTableRow) = "</td>\n<td>"; 
P(endTableRow) = "</td>\n</tr>\n";
P(endTableandForm) = "</table>\n</form>\n";
P(clickPos) = "Click on the position for device:";
P(grayColorBegin) = "<font color='#AAAAAA'>";
P(strEndForm) = "</form>";
P(strEndTable) = "</table>";
P(actionOnOff)="<center><select name='x'> <option value='1'>On<option value='0'>Off</select></center>";
P(actionUpDown)="<center><select name='x'> <option value='1'>Open<option value='0'>Close</select></center>";
P(htmlCommonEnding) = "</form></body></html>";
P(htmlCommonHeader) =
        "<html>\n"
        "<meta name='viewport' content='initial-scale=0.85, user-scalable=no, width=device-width, height=device-height'>\n"
        "<head><title>Ardomotic Server</title></head>\n"
        "<link rel='icon' href='1.gif'>\n"
        "<link rel='apple-touch-icon' href='1.gif'/>\n";


// Adds an images to the html page being constructed
// is_input -> marks the image as an input for an action
// img_name -> the name is not the filename of the image, but the name that will be used
//             when clicking the image to submit a POST
// name_number -> this number is added to img_name in certain cases. for instance, the 
//                devices then we can turn on or off by clicking are called d1, d2, d3,...
// file_number -> number used for the filename (1.gif, 2.gif, ...)
// image_width, int image_height -> dimensions for the image to be displayed
void htmlAddImage(boolean is_input, char img_name, byte name_number, byte file_number, int image_width, int image_height)
{
    SP(strCenterBegin);
    if (is_input) { SPRINT(strImgInput, "<input type='image' "); }
    else { SPRINT(strImgSimple, "<img "); }
    SPRINT(strImgName, "name='");
    webserver << img_name;
    if (name_number<IMAGE_NO_NUMBER)  
        webserver.print(name_number, DEC);
    SPRINT(strImgSrc, "' src='");
    // To keep the images on an external site, simply uncomment the following line
    // and replace the URL with the correct address. Note that the images must still
    // be named 1.gif, 2.gif, 3.gif, .....  
    //SPRINT(strImgUrl, "http://my.image.site.com/");
   
    itoa(file_number, buff, 10);
    webserver << buff;
    SPRINT(strFileGIF, ".gif");    
    SPRINT(strImgStyle, "' style='");
    serverPrintDimensions(image_width, image_height);
    SPRINT(strImgEnd, "'>");
}

void htmlAddHeaderImage(char* image_name, byte image_number)
{
    SPRINT(strHeaderImgStart1, "<td><center><a href='");  
    webserver << image_name;
    SPRINT(strHeaderImgStart2, "'>"); 
    htmlAddImage(false, 'z', IMAGE_NO_NUMBER, image_number, 80, 80);
    SPRINT(strHeaderImgStart3, "</td>\n"); 
}

void htmlAddForm(char* image_name)
{
    SPRINT(strFormStart1, "<form name='input' action='");
    webserver << image_name;   
    SPRINT(strFormStart2, "' method='post'>\n");
}

void htmlBeginning()
{
    SP(htmlCommonHeader);
    SPRINT(htmlBegin1, "<script type='text/javascript'>\n"
        "var a=");
    webserver << getDaySeconds();
    SPRINT(htmlBegin2, 
        "\nfunction dt(){\n"
        "h=Math.floor(a/3600)\n"
        "h=(h<10?'0':'')+h;\n"
        "m=Math.floor((a%3600)/60)\n"
        "m=(m<10?'0':'')+m;\n"
        "s=Math.floor(a%60)\n"
        "s=(s<10?'0':'')+s;\n"
        "document.getElementById('st').innerHTML=h+':'+m+':'+s\n"
        "a = a+1}\n"
        "window.onload=function(){dt();setInterval('dt()', 1000)}\n"
        "</script>\n"
        "</head>\n"
        "<body>\n"
        "<center>");
        htmlAddForm(pageNames[PAGE_NUMBER_LOGIN]);              
        SPRINT(htmlBegin3, 
        "<table border='0' cellpadding='0' cellspacing='0' width = '280' border = '0'>\n"
        "<tr><td><center><b>Ardomotic v1.0b</b></center></td><td>");
        htmlAddImage(true, PARAM_LOGOUT, IMAGE_NO_NUMBER, IMAGE_POWER, 30, 30);   
        SPRINT(htmlBegin4, 
        "</tr>"
        "</table>"
        "<table border='0' cellpadding='0' cellspacing='0' width = '280' border = '0'>\n"
        "<tr>\n");
        htmlAddHeaderImage(pageNames[PAGE_NUMBER_DIRECT], 2);
        htmlAddHeaderImage(pageNames[PAGE_NUMBER_SCHEDULER], 3);
        htmlAddHeaderImage(pageNames[PAGE_NUMBER_BLUEPRINT], 4);
        SPRINT(htmlBegin6, 
        "</tr>\n"
        "</table>\n"
        "</form>");
}

void serverPrintDimensions(int x, int y)
{
    SPRINT(strWidth, " width:");
    webserver.print(x, DEC);
    SPRINT(strHeight, "; height:");
    webserver.print(y, DEC);
}

boolean processLogin(WebServer &server)
{
    #ifdef DEBUG_POST
    Serial.println("Processing login POST...");
    #endif
    while( server.readPOSTparam( name, NAMELEN, value, VALUELEN))
    {
        #ifdef DEBUG_POST
        Serial.print("name=\"");
        Serial.print(name);
        Serial.print("\"  value=\"");
        Serial.print(value);
        Serial.println("\"");
        #endif
        if (name[0] == PARAM_LOGOUT )
        {
            webserver.clearCookie(millis());
            break;
        }
        if (name[0] == PARAM_PASSWORD )
        {
            if (strcmp(value, SYSTEM_PASSWORD) == 0)
            {
                #ifdef DEBUG_POST
                Serial.println("ACCEPTED!!! :)");
                #endif
                server.sendCookie(millis());
                return true;
            }
            #ifdef DEBUG_POST
            else 
                Serial.println("Denied!!! :(");
            #endif
        }
    }
    return false;
}

void processPost(WebServer &server)
{
    #ifdef DEBUG_RAM
    printRam();
    #endif
    int xpos = -1;
    int ypos = -1;
    
    byte hours, minutes, seconds;
    long action_time = -1;
    byte action_data_1 = -1;
    byte action_device = -1;
    byte action_number = -1;

    #ifdef DEBUG_POST
    Serial.println("Processing POST...");
    #endif
    while( server.readPOSTparam( name, NAMELEN, value, VALUELEN))
    {
        #ifdef DEBUG_POST
        Serial.print("name=");
        Serial.print(name);
        Serial.print("  value=");
        Serial.println(value);
        #endif
        byte param_number = atoi(name+1);
        if (name[1] == '.')
        {
            if (name[2]=='x') xpos = atoi(value);
            if (name[2]=='y') ypos = atoi(value);
        }
        // If the parameter refers to a device, sets the new status of that device
        if (name[0] == PARAM_DEVICE && param_number>=1 && param_number <= devices.count())
        {
            setDeviceStatus(param_number, !getDeviceStatus(param_number));
            return;
        }
        // Resets all of the device positions on the blueprint view
        if (name[0] == PARAM_RESET )
        {
            resetDevicePositions();
            return;
        }
        // Sets the position of the next device on the blueprint view
        if (name[0] == PARAM_MAP && xpos>-1 && ypos>-1)
        {
            byte unpositioned_device = nextDeviceWithNoPosition();
            if (unpositioned_device > 0)
                setDevicePosition(unpositioned_device, (float)xpos/BLUEPRINT_WIDTH, (float)ypos/BLUEPRINT_HEIGHT);            
            return;
        }
        // Time of the schedule to add
        if (name[0] == PARAM_TIME)
        {
            getTimeFromString(value, &hours, &minutes, &seconds);
            action_time = timeToSeconds( &hours, &minutes, &seconds);
        }
        // Number of the device to add the scheduled action
        if (name[0] == PARAM_ADD)
        {
            action_device = atoi(name+1);
        }
        // Details on the action to be scheduled
        if (name[0] == PARAM_DATA_1 )
        {
            action_data_1 = atoi(value);
        }
        // Save the data from all the schedules on the Flash memory
        if (name[0] == PARAM_SAVE )
        {
            saveSchedules();
            return;
        }
        // Order to remove a schedurle action from the list
        if (name[0] == PARAM_REMOVE && param_number >=1 && param_number <= schedules.count())
        {
            removeSchedule(param_number);
            return;
        }
        if (name[0] == PARAM_CLOCK )
        {
            getTimeFromString(value, &hours, &minutes, &seconds);
            // Only the time is used so far, so the date values are irrelevant
            setClock(hours, minutes, seconds, 1, 1, 2012, 1);
            return;
        }
    }
    // Checks if there is valid data to create a new action
    // Time exists and is valid
    if ( action_time > -1 && action_time <CLOCK_MAX_DAY_SECONDS &&
         action_device > 0 && action_device <= devices.count() )
    {
        addSchedule(action_time, action_device, action_data_1);
    }
}

// Page for the scheduler configuration
void schedulerCmd(WebServer &server, WebServer::ConnectionType type, char *url_tail, bool tail_complete)
{
    long action_time = -1;
    #ifdef DEBUG_PAGES
    Serial.println("# SCHEDULER #");
    #endif
    processPost(server);
    server.httpSuccess();
    htmlBeginning();
    SPRINT(beginListScheduledActions1, 
        "<b>List of scheduled actions</b>\n");        
    htmlAddForm(pageNames[PAGE_NUMBER_SCHEDULER]);
    SPRINT(beginListScheduledActions2, 
        "<table border='1' width = '350'>\n"
        "<tr>\n"
        "<th>Time</th><th>Device</th><th>Action</th><th>Remove</th>\n</tr>\n");   
    long  time_now = getDaySeconds();
    // Creates the table of all the scheduled actions
    for (byte i=1; i<=schedules.count(); i++)
    {
        SP(beginTableRow);
        action_time = getActionTime(i);
        if (time_now>action_time) SP(grayColorBegin);
        timeString(buff, &action_time);
        server << buff;
        SP(middleTableRow);
        byte act_dev = getScheduleDevice(i);
        if (time_now>action_time) SP(grayColorBegin);
        SP( getDeviceName(act_dev));
        SP(middleTableRow);
        htmlAddImage(false, PARAM_DEVICE, i, getDeviceImage(act_dev, getScheduleCommand(i)) , DEVICE_IMAGE_DIMENSION, DEVICE_IMAGE_DIMENSION);
        SP(middleTableRow);
        htmlAddImage(true, PARAM_REMOVE, i, IMAGE_CANCEL , DEVICE_IMAGE_DIMENSION, DEVICE_IMAGE_DIMENSION);
        SP(endTableRow);
    }
    SP(strEndTable);
    htmlAddImage(true, PARAM_SAVE, IMAGE_NO_NUMBER , IMAGE_SAVE , 50, 50);
    SP(strEndForm);
    SPRINT(addNewEvents1, 
        "<br/>\n"
        "<b>Schedule new actions</b><br/>\n"
        "<table border='1' width = '350'>\n"
        "<tr>\n"
        "<th>Time</th><th>Device</th><th>Action</th><th>Add</th>\n</tr>\n");   
    char timebuff[10];
    timeString(timebuff); 
    // Creates the table of all the devices, allowing to schedule new actions
    for (byte i=1; i<=devices.count(); i++)
    {
        htmlAddForm(pageNames[PAGE_NUMBER_SCHEDULER]);

        SP(beginTableRow);
        SPRINT(timeInput1, "<center><input type='text' size='7' name='t' value='");
        server << timebuff;
        SPRINT(timeInput2, "'><center>");
        SP(middleTableRow);    
        SP(getDeviceName(i));
        SP(middleTableRow);
        switch (getDeviceType(i))
        {
            case DEVICE_TYPE_SWITCH: SP(actionOnOff); break;
            case DEVICE_TYPE_SLIDER: SP(actionUpDown); break;
        }
        SP(middleTableRow);
        htmlAddImage(true, PARAM_ADD, i, IMAGE_ADD , DEVICE_IMAGE_DIMENSION, DEVICE_IMAGE_DIMENSION);
        SP(endTableRow);
        SPRINT(endForm, "</form>");
    }
    SP(strEndTable);
    SPRINT(strClock1, "<b><center>Current Server Time:</center></b><span id='st'></span></form>"
                     "<br/>"
                     "<b>Set server time</b>"
                     "<table>");

    htmlAddForm(pageNames[PAGE_NUMBER_SCHEDULER]);
                     
    SPRINT(strClock2, "<td><center><input type='text' size='7' name='c' value='12:00:00'><td>");
    htmlAddImage(true, PARAM_TIME, 0, IMAGE_CONFIRM , DEVICE_IMAGE_DIMENSION, DEVICE_IMAGE_DIMENSION);
    SP(htmlCommonEnding);
}

void directCmd(WebServer &server, WebServer::ConnectionType type, char *url_tail, bool tail_complete)
{  
    #ifdef DEBUG_PAGES
    Serial.println("# DIRECT #");
    #endif

    processPost(server);
    server.httpSuccess(1, false);
    
    htmlBeginning();
    SPRINT(beginFormHomepage1,
    "<b>Devices in the system</b>\n");
    htmlAddForm(pageNames[PAGE_NUMBER_DIRECT]);
    SPRINT(beginFormHomepage2,
        "<center><table border='1' width = '280'>\n"
        "<tr>\n"
        "<th>Status</th><th>Component</th>\n</tr>\n");
    // Creates the table of all the devices in the system for direct control
    for (byte i=1; i<=devices.count(); i++)
    {
        SP(beginTableRow);
        htmlAddImage(true, PARAM_DEVICE, i, getDeviceImage(i, getDeviceStatus(i)) , DEVICE_IMAGE_DIMENSION, DEVICE_IMAGE_DIMENSION);
        SP(middleTableRow);
        SP(getDeviceName(i));
        SP(endTableRow);
    }
    SP(endTableandForm);
    SP(htmlCommonEnding);
}

void blueprintCmd(WebServer &server, WebServer::ConnectionType type, char *url_tail, bool tail_complete)
{  
    #ifdef DEBUG_PAGES
    Serial.println("# BLUEPRINT #");
    #endif
    processPost(server);
    server.httpSuccess();
    htmlBeginning();

    byte unpositioned_device = nextDeviceWithNoPosition();
    if (unpositioned_device > 0)
    {
        SP(clickPos);
        SP(getDeviceName(unpositioned_device));
        SP(newLine);
    }
    SPRINT(beginDrawPlant1,
        "<center>\n"
        "<b>Blueprint of the house</b>\n");
    htmlAddForm(pageNames[PAGE_NUMBER_BLUEPRINT]);        
    SPRINT(beginDrawPlant2,
        "<div style='position:relative;");
    serverPrintDimensions(BLUEPRINT_WIDTH, BLUEPRINT_HEIGHT);
    SPRINT(beginDrawPlant3,
        "'>\n"
	"<div style='position:absolute; top:0; left:0'>");
    htmlAddImage((unpositioned_device > 0), PARAM_MAP, IMAGE_NO_NUMBER, IMAGE_PLANT, BLUEPRINT_WIDTH, BLUEPRINT_HEIGHT);

    SPRINT(strEndDiv, "</div>\n");
    for (byte i=1; i<=devices.count(); i++)
    {
        float xpos = getDeviceXPos(i) * BLUEPRINT_WIDTH - DEVICE_IMAGE_DIMENSION/2;
        float ypos = getDeviceYPos(i) * BLUEPRINT_HEIGHT - DEVICE_IMAGE_DIMENSION/2;
        if (xpos>=0 || ypos>=0)
        { 
//            Serial.print("Device position:  "); Serial.print("X:"); Serial.print(xpos); Serial.print("  Y:"); Serial.println(ypos);       
            byte dev_status = getDeviceStatus( i );
            SPRINT(plantDivBegin, "<div style='position:absolute; top:");
            server.print(ypos, DEC);
            SPRINT(plantDivMiddle1, "; left:");
            server.print(xpos, DEC);
            SPRINT(plantDivMiddle2, "'>");
            htmlAddImage(true, PARAM_DEVICE, i, getDeviceImage(i, getDeviceStatus(i)), DEVICE_IMAGE_DIMENSION, DEVICE_IMAGE_DIMENSION); 
            SP(strEndDiv);
        }
    }
    SPRINT(endDrawPlant, "</div>\n");
    SPRINT(resetButton, "Reset All Positions<br>\n");
    htmlAddImage(true, PARAM_RESET, IMAGE_NO_NUMBER, IMAGE_CANCEL, DEVICE_IMAGE_DIMENSION, DEVICE_IMAGE_DIMENSION);            
    SP(htmlCommonEnding);
}

void loginCmd(WebServer &server, WebServer::ConnectionType type, char *url_tail, bool tail_complete)
{  
    #ifdef DEBUG_PAGES
    Serial.println("# LOGIN #");
    #endif
    
    // processPost returns true if the user typed the right password
    // so when this happens, httpSucess sends the new Cookie in the headers
    boolean login_sucess = processLogin(server);

    // If the login was sucesseful, redirects to the desired page
    if (login_sucess)
    {
        blueprintCmd(server, type, url_tail, tail_complete);
    }
    // Otherwise keeps showing the login page
    else
    {
        server.httpSuccess(1, false);
        SP(htmlCommonHeader);
        SPRINT( htmlLoginPage1, 
        "</head>\n\n"
        "<body>\n"
        "<center>");
        htmlAddImage(false, 'z', IMAGE_NO_NUMBER, 1, 80, 80);
        SPRINT( htmlLoginPage2, 
        "</a>\n"
        "<br><b>Ardomotics Login</b>\n");
        htmlAddForm(pageNames[PAGE_NUMBER_DIRECT]);                
        SPRINT( htmlLoginPage3, 
        "<input type='password' size='12' name='p'>\n");
        SP(htmlCommonEnding);
    }
}

#define IMAGE_BYTE_BUFFER 128

boolean sendFileCommand( WebServer &server, WebServer::ConnectionType type, char *filename)
{
    int file_number;
    
    strcpy(buff, filename);
    if (buff[1] == '.') buff[1] = 0;
    buff[2] = 0;
    
    file_number = atoi(buff);
    // Returns false if the image was not found in the list
    if (file_number == 0 || file_number > TOTAL_IMAGES) return false;

    file_number--;
    server.httpSuccess(CONTENT_IMAGE_GIF, true);
    if (type == WebServer::GET)
    {
        unsigned int image_address = flash.readInt(1+2*file_number);
        byte bytebuff[IMAGE_BYTE_BUFFER];
//        unsigned int size_remaining = flash.readInt(image_addresses[file_number]);
        // The first two bytes contain the size of the image
        unsigned int size_remaining = flash.readInt(image_address);

        #ifdef DEBUG_PAGES
        Serial.print("File:");
        Serial.print(file_number);
        Serial.print("   Size:");
        Serial.println(size_remaining, DEC);
        int bytes_sent = 0;
        #endif

        // The actual image data starts here
        image_address += 2;
        while (size_remaining > 0)
        {
            int next_chunk_size = (size_remaining > IMAGE_BYTE_BUFFER ? IMAGE_BYTE_BUFFER : size_remaining);
            
            flash.readBlock( image_address, bytebuff, IMAGE_BYTE_BUFFER );           
            image_address += IMAGE_BYTE_BUFFER;
            server.write(bytebuff, next_chunk_size );
            size_remaining -= next_chunk_size;
            
            #ifdef DEBUG_MODE>1
            bytes_sent += next_chunk_size;
            #endif
        }
    }
}

void setup()
{
    // Initializing all the hardware...
    Serial.begin(115200);   
    Serial.print("Starting...");
    Wire.begin();
    sr.initRegisters(LOW);
    sr.writeRegisters();

//    setClock(23, 59, 40, 6, 2, 12, 3);
    long now = getDaySeconds();
    webserver = WebServer(PREFIX, 80, now);
    Ethernet.begin(mac, ip);
    webserver.begin();

    // Adding all the devices to the system list
    ADD_DEVICE_SWITCH( "Hall Light", DEVICE_CONTROL_CHACON, IMAGE_LIGHT_ON,  0, 4);
    ADD_DEVICE_SWITCH( "Kitchen Light", DEVICE_CONTROL_CHACON, IMAGE_LIGHT_ON, 0, 3);
    ADD_DEVICE_SWITCH( "Utilities Room Light", DEVICE_CONTROL_CHACON, IMAGE_LIGHT_ON, 0, 7);
    ADD_DEVICE_SWITCH( "Pantry Light", DEVICE_CONTROL_DIRECT_PIN, IMAGE_LIGHT_ON, 0, 8);
    ADD_DEVICE_SWITCH( "Bathroom Light", DEVICE_CONTROL_SHIFT_REGISTER, IMAGE_LIGHT_ON, 0, 5);
    ADD_DEVICE_SWITCH( "Living Room Light", DEVICE_CONTROL_CHACON, IMAGE_LIGHT_ON, 0, 0);
    ADD_DEVICE_SWITCH( "Bedroom Light", DEVICE_CONTROL_CHACON, IMAGE_LIGHT_ON, 0, 2);
    ADD_DEVICE_SWITCH( "Office Light", DEVICE_CONTROL_CHACON, IMAGE_LIGHT_ON, 0, 1);

//    ADD_DEVICE_SWITCH( "Desk Power", DEVICE_CONTROL_CHACON, IMAGE_OUTLET_ON, 0, 12);
    ADD_DEVICE_SWITCH( "Desk Power", DEVICE_CONTROL_CHACON, IMAGE_OUTLET_ON, 0, 15);
    
    ADD_DEVICE_SLIDER( "Living Room Window 1", DEVICE_CONTROL_SHIFT_REGISTER, IMAGE_WINDOW_OPEN, 0, 3, 2, 3);
    ADD_DEVICE_SLIDER( "Living Room Window 2", DEVICE_CONTROL_SHIFT_REGISTER, IMAGE_WINDOW_OPEN, 0, 5, 4, 3);
    ADD_DEVICE_SLIDER( "Office Window", DEVICE_CONTROL_SHIFT_REGISTER, IMAGE_WINDOW_OPEN, 0, 0, 1, 3);

    ADD_DEVICE_TIMER( "Building Door", DEVICE_CONTROL_CHACON, IMAGE_KEY_ON, 0, 208, 0, 1);

/*
    // Adding all the devices to the system list
    ADD_DEVICE_SWITCH( "Luz Hall", DEVICE_CONTROL_CHACON, IMAGE_LIGHT_ON,  0, 4);
    ADD_DEVICE_SWITCH( "Luz Cozinha", DEVICE_CONTROL_CHACON, IMAGE_LIGHT_ON, 0, 3);
    ADD_DEVICE_SWITCH( "Luz Sala Máquinas", DEVICE_CONTROL_CHACON, IMAGE_LIGHT_ON, 0, 7);
    ADD_DEVICE_SWITCH( "Luz Despensa", DEVICE_CONTROL_WIRED, IMAGE_LIGHT_ON, 0, 6);
    ADD_DEVICE_SWITCH( "Luz Casa Banho", DEVICE_CONTROL_WIRED, IMAGE_LIGHT_ON, 0, 5);
    ADD_DEVICE_SWITCH( "Luz Sala", DEVICE_CONTROL_CHACON, IMAGE_LIGHT_ON, 0, 0);
    ADD_DEVICE_SWITCH( "Luz Quarto", DEVICE_CONTROL_CHACON, IMAGE_LIGHT_ON, 0, 2);
    ADD_DEVICE_SWITCH( "Luz Escritório", DEVICE_CONTROL_CHACON, IMAGE_LIGHT_ON, 0, 1);

    ADD_DEVICE_SWITCH( "Ficha Computador", DEVICE_CONTROL_CHACON, IMAGE_OUTLET_ON, 0, 12);
    
    ADD_DEVICE_SLIDER( "Janela Sala 1", DEVICE_CONTROL_WIRED, IMAGE_WINDOW_OPEN, 0, 3, 2, 18);
    ADD_DEVICE_SLIDER( "Janela Sala 2", DEVICE_CONTROL_WIRED, IMAGE_WINDOW_OPEN, 0, 5, 4, 18);
    ADD_DEVICE_SLIDER( "Janela Escritório", DEVICE_CONTROL_WIRED, IMAGE_WINDOW_OPEN, 0, 0, 1, 18);

    ADD_DEVICE_TIMER( "Building Door", DEVICE_CONTROL_CHACON, IMAGE_KEY, 0, 208, 0, 1);
*/
    // Do NOT load data the first time the code runs, or after adding or removing devices
    // These 2 functions will read data from the Arduino flash memory, and if the memory 
    // hasn't been written before, strange things may happen
//    loadDevicePositions();
//    loadSchedules();

    // Configures the HTML pages that the server will respond to
    // The default command and login command should be the login page
    webserver.setDefaultCommand(&loginCmd);
    webserver.setLoginCommand(&loginCmd);
    // FileCommands is functions that responds to binary file requests (the images)
    webserver.setFileCommand(&sendFileCommand);
    webserver.addCommand(pageNames[PAGE_NUMBER_DIRECT], &directCmd);
    webserver.addCommand(pageNames[PAGE_NUMBER_SCHEDULER], &schedulerCmd);
    webserver.addCommand(pageNames[PAGE_NUMBER_BLUEPRINT], &blueprintCmd);
    webserver.addCommand(pageNames[PAGE_NUMBER_LOGIN], &loginCmd);
    Serial.println(" done!");
}

// Variables used to detect the moment the day changes, in order to 
// reset the scheduler and adjust the event time values
long previous_time, time;
void loop()
{
    // Process incoming connections one at a time forever
    webserver.processConnection();

    // Checks if there are any new events or actions to be processed    
    previous_time = time;
    time = getDaySeconds();
    
    processSchedules(previous_time, time);
    processEvents(previous_time, time);
}
