#ifndef _MY_UTILS_H
#define _MY_UTILS_H

// General utilites for debug purposes and to simplify parts of the code

#include <WProgram.h>
#include <EEPROM.h>

#define SPRINT(a,b) P(a) = b; webserver.printP(a)
#define SP(a) webserver.printP(a)

#define PRINT(token) Serial.print(#token ":"); Serial.print(token); Serial.print("   ");
#define PRINT_DEC(token) Serial.print(#token ":"); Serial.print(token, DEC); Serial.print("   ");

template <class T> int EEPROM_writeAnything(int ee, const T& value)
{
    const byte* p = (const byte*)(const void*)&value;
    int i;
    for (i = 0; i < sizeof(value); i++)
          EEPROM.write(ee++, *p++);
    return i;
}

template <class T> int EEPROM_readAnything(int ee, T& value)
{
    byte* p = (byte*)(void*)&value;
    int i;
    for (i = 0; i < sizeof(value); i++)
          *p++ = EEPROM.read(ee++);
    return i;
}

template<class T>
inline Print &operator <<(Print &obj, T arg)
{ 
    obj.print(arg); 
    return obj; 
}

extern struct __freelist *__flp;
extern uint8_t* __brkval;

// checks the amount of free ram
int freeRam ()
{
    extern int __heap_start;
    //, *__brkval;
    int v;
    return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval);
}

int float2Int(float x)
{
    return (int)(x*32767);
}

float int2Float(int x)
{
    return (float)x/32767.0;
}

#ifdef DEBUG_RAM
void printRam()
{
    Serial.print("Free RAM:");
    Serial.println(freeRam());
}
#endif

#if DEBUG_MODE>1
void printTime(byte hours, byte minutes, byte seconds)
{
    Serial.print(hours, DEC);
    Serial.print(':');
    Serial.print(minutes, DEC);
    Serial.print(':');
    Serial.println(seconds, DEC);
}
#endif


#endif
